==> mutual_credit-5.0.9.tar.gz: mutual_credit/src/Element/WorthForm.php <==
<?php

namespace Drupal\mcapi\Element;

use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Mcapi;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\FormElement;

/**
 * Provides a worth field form element.
 *
 * @FormElement("worth_form")
 */
class WorthForm extends FormElement {
  use \Drupal\Core\Render\Element\CompositeFormElementTrait;

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#process' => [
        [$class, 'process'],
      ],
      '#element_validate' => [
        [$class, 'validate'],
      ],
      '#pre_render' => [
        [$class, 'preRenderCompositeFormElement'],
      ],
      '#attached' => ['library' => ['mcapi/mcapi.worth.element']],
      '#attributes' => ['class' => ['worth']],
      '#minus' => FALSE, // TRUE if negative values allowed
      '#config' => FALSE,
      '#max' => NULL,
      '#allowed_curr_ids' => [],
      '#oneofmany' => FALSE,
      '#theme_wrappers' => ['form_element'],
    ];
  }

  /**
   * Form element processor callback for 'worth'.
   *
   * Takes the raw #default_values and makes child widgets according to the
   * currency format receives something like this:
   * [
   * '#type' => 'worth_form',
   * '#default_value' => [
   *   'curr_id',
   *   'value'
   * ],
   * '#allowed_curr_ids' => [cc],
   * '#config' => $element['#config'],
   * '#max' => NULL,
   * '#min' => 0,
   * ];.
   */
  public static function process($element, FormStateInterface $form_state, $form) {
    if (empty($element['#allowed_curr_ids'])) {
      $element['#allowed_curr_ids'] = array_keys(Currency::loadMultiple());
    }
    // We might need to filter #default_value not in config mode.
    if (count($element['#allowed_curr_ids']) == 1) {
      $currency = Currency::load(reset($element['#allowed_curr_ids']));
      if (!$currency) {
        trigger_error('Problem with allowed_currency_ids in WorthForm', E_USER_ERROR);
      }
      Self::subfields($element, $currency);
    }
    // Multiple choice of currencies, but only showing one widget.
    else {
      $element['curr_id'] = [
        '#type' => 'select',
        '#options' => array_intersect_key(
          Mcapi::entityOptions('mcapi_currency'),
          Currency::loadMultiple($element['#allowed_curr_ids'] ?: NULL)
        ),
        '#required' => TRUE,
      ];
      $element['value'] = [
        '#type' => 'number',
        '#default_value' => $element['#default_value'],
      ];
      if ($element['#config']) {
        $element['value']['#size'] = 10;
        $element['value']['#maxlength'] = 10;
      }
    }
    return $element;
  }

  /**
   * Build the widget for a single currency.
   *
   * @param array $element
   *   An unprocessed part of a render array.
   * @param Currency $currency
   *   A currency entity.
   */
  public static function subfields(array &$element, Currency $currency) {
    $value = $element['#default_value'];
    if ($element['#config'] && (!is_numeric($value) || $value == 0)) {
      $value_parts = [];
    }
    else {
      $value_parts = (strlen($value)) ?
        $currency->formattedParts(intval($value)) :
        [];
    }
    $element['curr_id'] = [
      '#type' => 'hidden',
      '#value'  => $currency->id(),
    ];
    foreach ($currency->format as $i => $component) {
      $element[$i] = ['#weight' => $i];
      // An odd number so render a form element.
      if ($i % 2) {
        $step = 1;
        $options = [];
        // We need to make a dropdown.
        if (strpos($component, '/')) {
          list($component, $divisor) = explode('/', $component);
          $base = $component + 1;
          $step = $base / $divisor;
          for ($j = 0; $j < $divisor; $j++) {
            $v = $j * $step;
            $options[intval($v)] = $v;
          }
        }
        $element[$i] += [
          '#default_value' => @$value_parts[$i],
          '#theme_wrappers' => [],
        ];
        // If a preset value isn't in the $options
        // then we ignore the options and use the numeric sub-widget.
        if (isset($options) && array_key_exists(intval(@$value_parts[$i]), $options)) {
          $element[$i] += [
            '#type' => 'select',
            '#options' => $options,
            '#attributes' => ['style' => 'width: inherit;'],
          ];
        }
        else {
          $size = strlen($component);
          $element[$i] += [
            '#type' => 'number',
            '#min' => $element['#minus'] ? NULL : 0,
            '#step' => $step,
            // #size has no effect in opera.
            '#size' => $size + ($element['#minus'] ? 1 : 0),
            '#maxlength' => $size + ($element['#minus'] ? 1 : 0),
            '#attributes' => ['style' => 'width:' . ($size + 1) . 'em;'],
            '#max' => $component//which is already all the nines
          ];

          if ($element['#config']) {
            // placeholder's only work in config fields.
            $element['#type'] = 'textfield';
            if (isset($element['#placeholder'])) {
              $placeholder_val = @$element['#placeholder'];
              $p_parts = $currency->formattedParts(abs(intval($placeholder_val)));
              $element[$i]['#placeholder'] = $p_parts[$i];
            }
          }
          else {
            // Leave the main value field empty & put zero placeholders in other
            // fields.
            $element[$i]['#placeholder'] = $i == 2 ? '' : str_pad('', $size, '0');
          }
        }
        unset($options);
      }
      // An even number so render it as markup.
      else {
        $element[$i]['#markup'] = $component;
      }
    }
    $cardinality = $element['#oneofmany'] ? 'multiple' : 'single';
    $element['#prefix'] = "<span class = \"worth $cardinality\">";
    $element['#suffix'] = "</span>";
  }

  /**
   * {@inheritdoc}
   *
   * @todo Stop the values of the child elements overwriting the $output here?
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input == FALSE or !isset($input['curr_id'])) {
      return $element['#default_value'];
    }
    return static::worthValue($input, $element['#config'], $element['#default_value']);
  }


  /**
   * Validate callback.
   */
  public static function validate(&$element, FormStateInterface $form_state) {
    $val = $element['#value'];
    if (empty($val)) {
      return;
    }
    if ($element['#config']) {
      // Test the formula.
      if (empty($val['value'])) {
        return;
      }
      $result = Self::calc($val['value'], 100);
      if (!is_numeric($result)) {
        $form_state->setError($element, t('Invalid formula'));
      }
    }
    else {
      // Check for allowed zero values.
      // zero values are only accepted if the currency allows and if cardinality
      // there is only one currency in the field.
      if ($val['value']) {
        if ($val['value'] < 0 && !$element['#minus']) {
          // This should never happen.
          $form_state->setError(
            $element['#value']['value'],
            t(
              'Negative amounts not allowed: !val',
              ['%val' => Currency::load($val['curr_id'])->format($element['#value']['value'])]
            )
          );
        }
      }
    }
  }

  /**
   * Calculate a transaction quantity based on a provided formala and quantity.
   *
   * @param string $formula
   *   Formula using [q] for base_quant. If it is just a number the number is
   *   returned as is. Otherwise [q]% should work or other variables to be
   *   determined.
   * @param int $base_value
   *   The starting value.
   *
   * @return int | NULL
   *   The calculated result.
   */
  public static function calc($formula, $base_value) {
    if (is_null($base_value)) {
      return 0;
    }
    if (is_numeric($formula)) {
      return $formula;
    }
    $proportion = str_replace('%', '', $formula);
    if (empty($base_value)) {
      // @todo there is a problem this variable is not used...
      $base_quant = 0;
    }
    if (is_numeric($proportion)) {
      return $base_value * $proportion / 100;
    }
    // $formula = str_replace('/', '//', $formula);.
    $equation = str_replace('[q]', $base_value, $formula) . ';';
    $val = eval('return ' . $equation);
    if (is_numeric($val)) {
      return $val;
    }
    $this->messenger()->addStatus(t('Problem with calculation for dependent transaction: @val', array('@val' => $val)));
  }

  /**
  *
  * @param array $input
  * @param type $config
  * @param type $default
  * @return type
  */
  static function worthValue(array $input, $config, $default) {
    if ($config && empty($input[1]) && empty($input[3])) {
      return [];
    }
    $val = Currency::load($input['curr_id'])->unformat($input);
    return [
      'curr_id' => $input['curr_id'],
      'value' => $val,
    ];
  }

}

==> mutual_credit-5.0.10.tar.gz: mutual_credit/src/Element/WorthForm.php <==
<?php

namespace Drupal\mcapi\Element;

use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Mcapi;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\FormElement;

/**
 * Provides a worth field form element.
 *
 * @FormElement("worth_form")
 */
class WorthForm extends FormElement {
  use \Drupal\Core\Render\Element\CompositeFormElementTrait;

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#process' => [
        [$class, 'process'],
      ],
      '#element_validate' => [
        [$class, 'validate'],
      ],
      '#pre_render' => [
        //[$class, 'preRenderCompositeFormElement'],
      ],
      '#attached' => ['library' => ['mcapi/mcapi.worth.element']],
      '#attributes' => ['class' => ['worth']],
      '#config' => FALSE,
      '#minus' => FALSE, // TRUE if negative values allowed (when #config is TRUE only)
      '#tree' => TRUE,
      '#allowed_curr_ids' => array_keys(Currency::loadMultiple()),
      '#oneofmany' => FALSE,
      '#theme_wrappers' => ['form_element'],
    ];
  }

  /**
   * Form element processor callback for 'worth'.
   *
   * Takes the raw #default_values and makes child widgets according to the
   * currency format receives something like this:
   * [
   * '#type' => 'worth_form',
   * '#default_value' => [
   *   'curr_id',
   *   'value'
   * ],
   * '#allowed_curr_ids' => [cc],
   * '#config' => $element['#config'],
   * '#max' => NULL,
   * '#min' => 0,
   * ];.
   */
  public static function process($element, FormStateInterface $form_state, $form) {
    // Render the widget for a known currency
    if (count($element['#allowed_curr_ids']) == 1) {
      $currency = Currency::load(reset($element['#allowed_curr_ids']));
      if (!$currency) {
        trigger_error('Problem with allowed_currency_ids in WorthForm', E_USER_ERROR);
      }
      Self::subfields($element, $currency);
    }
    // Render the widget for a selection of currencies.
    else {
      $element['curr_id'] = [
        '#type' => 'select',
        '#options' => array_intersect_key(
          Mcapi::entityOptions('mcapi_currency'),
          Currency::loadMultiple($element['#allowed_curr_ids'] ?: NULL)
        ),
        '#required' => TRUE,
      ];
      $element['value'] = [
        '#type' => 'number',
        '#default_value' => $element['#default_value'],
      ];
      if ($element['#config']) {
        $element['value']['#size'] = 10;
        $element['value']['#maxlength'] = 10;
      }
    }
    return $element;
  }

  /**
   * Build the widget for a single currency.
   *
   * @param array $element
   *   An unprocessed part of a render array.
   * @param Currency $currency
   *   A currency entity.
   */
  public static function subfields(array &$element, Currency $currency) {
    $element['curr_id'] = [
      '#type' => 'hidden',
      '#value'  => $currency->id(),
    ];
    $value = $element['#default_value'] ? $element['#default_value']['value'] : 0; // the raw value
    if ($element['#config'] && (!is_numeric($value) || $value == 0)) {
      $value_parts = [];
    }
    else {
      $value_parts = (strlen($value)) ?
        $currency->formattedParts($value) :
        [];
    }
    $max_chars = strlen($currency->format['main']);
    $dotpos = strrpos($currency->format['main'], '.');
    $element['main'] = [
      '#type' => 'number',
      '#default_value' => $value_parts['main']??0,
      '#min' => $element['#minus'] ? NULL : 0,
      '#step' => $dotpos ? pow(10, -$dotpos) : 1,
      '#size' => $max_chars + ($element['#minus'] ? 1 : 0),
      '#max' => $currency->format['main'], // Which is all the nines,
      '#min' => 0,
      '#field_prefix' => $currency->format['f1'],
      '#field_suffix' => $currency->format['f2']??'',
      '#weight' => 1,
    ];
    if (isset($currency->format['subdivision'])) {
      $subdivision = $currency->format['subdivision'];
      // We need to make a dropdown.
      if (strpos($subdivision, '/')) {
        $options = [];
        list($subdivision, $divisor) = explode('/', $subdivision);
        $base = $subdivision + 1;
        $step = $base / $divisor;
        for ($j = 0; $j < $divisor; $j++) {
          $v = $j * $step;
          $options[intval($v)] = $v;
        }
        // If a preset value isn't in the $options
        // then we ignore the options and use the numeric sub-widget.
        if (array_key_exists($value_parts['subdivision'], $options)) {
          $element['subdivision'] = [
            '#type' => 'select',
            '#options' => $options,
            '#attributes' => ['style' => 'width: inherit;'],
            '#weight' => 3
          ];
        }
        else {
          $size = strlen($subdivision);
          $element['subdivision'] = [
            '#type' => 'number',
            '#min' => $element['#minus'] ? NULL : 0,
            '#step' => $step,
            '#size' => $size,
            '#maxlength' => $size,
            '#attributes' => ['style' => 'width:' . ($size + 1) . 'en;'],
            '#max' => $subdivision //Which is already all the nines
          ];
          if ($element['#config']) {
            // Placeholders only work in config fields.
            $element['#type'] = 'textfield';
            if (isset($element['#placeholder'])) {
              $placeholder_val = @$element['#placeholder'];
              $p_parts = $currency->formattedParts(abs(intval($placeholder_val)));
              $element['main']['#placeholder'] = $p_parts['main'];
              if (isset($p_parts['subdivision'])) {
                $element['subdivision']['#placeholder'] = $p_parts['subdivision'];
              }
            }
          }
          else {
            // Leave the main value field empty & put zero placeholders in other
            // fields.
            $element['subdivision']['#placeholder'] = str_pad('', $size, '0');
          }
        }
      }
      else {
        $element['subdivision'] = [
          '#type' => 'number',
          '#default_value' => $value_parts['subdivision']??0,
          '#min' => 0,
          '#step' => 1,
          // #size has no effect in opera.
          '#size' => strlen($value_parts['subdivision']),
          '#maxlength' => strlen($value_parts['subdivision']),
          '#attributes' => ['style' => 'display:inline;'],
          '#max' => $currency->format['subdivision'], // E.G. 59 minutes
          '#weight' => 3
        ];
      }
      $element['subdivision'] += [
        '#weight' => 3,
        '#default_value' => $value_parts['subdivision']??0,
        '#theme_wrappers' => [],
      ];
    }
    if (isset($currency->format['f3'])) {
      $element['f3']['#markup'] = '<span>'.$currency->format['f3'].'</span>';
      $element['f3']['#weight'] = 5;
    }

    if ($element['#config'] and $element['#minus']) {
      $element['main']['#field_prefix'] = '-'.($element['main']['#field_prefix'] ??'');
      $element['main']['#default_value'] = -$element['main']['#default_value'];
    }

    $element['#attributes']['class'][] = $element['#oneofmany'] ? 'multiple' : 'single';
  }

  /**
   * {@inheritdoc}
   *
   * @todo Stop the values of the child elements overwriting the $output here?
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input == FALSE or !isset($input['curr_id'])) {
      return $element['#default_value'];
    }
    $result = static::worthValue($input, (bool)$element['#config']);
    if ($element['#config'] and $element['#minus']) {
      $result['value'] = -$result['value'];
    }
    return $result;
  }


  /**
   * Validate callback.
   */
  public static function validate(&$element, FormStateInterface $form_state) {
    $val = $element['#value'];
    if (empty($val)) {
      return;
    }
    if ($element['#config']) {
      // Test the formula.
      if (empty($val['value'])) {
        return;
      }
      $result = Self::calc($val['value'], 100);
      if (!is_numeric($result)) {
        $form_state->setError($element, t('Invalid formula'));
      }
    }
    else {
      // Check for allowed zero values.
      // zero values are only accepted if the currency allows and if cardinality
      // there is only one currency in the field.
      if ($val['value']) {
        if ($val['value'] < 0 && !($element['#config'] && $element['#minus'])) {
          // This should never happen.
          $form_state->setError(
            $element['#value']['value'],
            t(
              'Negative amounts not allowed: %val',
              ['%val' => Currency::load($val['curr_id'])->format($element['#value']['value'])]
            )
          );
        }
      }
    }
  }

  /**
   * Calculate a transaction quantity based on a provided formala and quantity.
   *
   * @param string $formula
   *   Formula using [q] for base_quant. If it is just a number the number is
   *   returned as is. Otherwise [q]% should work or other variables to be
   *   determined.
   * @param int $base_value
   *   The starting value.
   *
   * @return int | NULL
   *   The calculated result.
   */
  public static function calc(string $formula, int $base_value) {
    if (is_null($base_value)) {
      return 0;
    }
    if (is_numeric($formula)) {
      return $formula;
    }
    $proportion = str_replace('%', '', $formula);
    if (empty($base_value)) {
      // @todo there is a problem this variable is not used...
      $base_quant = 0;
    }
    if (is_numeric($proportion)) {
      return $base_value * $proportion / 100;
    }
    // $formula = str_replace('/', '//', $formula);.
    $equation = str_replace('[q]', $base_value, $formula) . ';';
    $val = eval('return ' . $equation);
    if (is_numeric($val)) {
      return $val;
    }
    $this->messenger()->addStatus(t('Problem with calculation for dependent transaction: @val', array('@val' => $val)));
  }

  /**
  *
  * @param array $input
  * @param bool $config
  * @return type
  */
  static function worthValue(array $input, bool $config) {
    if ($config && empty($input['main']) && empty($input['subdivision'])) {
      return [];
    }
    $val = Currency::load($input['curr_id'])->unformat($input);
    return [
      'curr_id' => $input['curr_id'],
      'value' => $val,
    ];
  }

}
